/* Simplethreads Instructional Thread Package
 * 
 * sthread_user.c - Implements the sthread API using user-level threads.
 *
 *    You need to impelement the routines in this file.
 *
 * Change Log:
 * 2002-04-15        rick
 *   - Initial version.
 */

#include <stdlib.h>
#include <assert.h>
#include <stdio.h>

#include <sthread.h>
#include <sthread_queue.h>
#include <sthread_user.h>
#include <sthread_ctx.h>

struct _sthread {
  sthread_ctx_t *saved_ctx;
  /* Add your fields to the thread data structure here */
/*<solution>*/  
  sthread_start_func_t start_routine;
  void *start_arg;
/*</solution>*/  
};

/*<solution>*/  
static sthread_queue_t ready_q;

static sthread_t current;
static sthread_t exiting;
static sthread_t main_th;

static void sthread_user_starter(void);

static void sthread_user_schedule(void);
static sthread_t sthread_user_current(void);
static sthread_t sthread_user_next(void);
static void sthread_user_check_cleanup(void);
static void sthread_user_cleanup(sthread_t sth);
static void sthread_user_ready(sthread_t ready_th);
/*</solution>*/

/*********************************************************************/
/* Part 1: Creating and Scheduling Threads                           */
/*********************************************************************/

void sthread_user_init(void) {
/*<solution>*/
  /* Create the main context */
  current = (sthread_t)malloc(sizeof(struct _sthread));
  assert(current != NULL);
   
  current->start_routine = NULL;
  current->start_arg = NULL;
  current->saved_ctx = (sthread_ctx_t*)malloc(sizeof(sthread_ctx_t));
  assert(current->saved_ctx != NULL);

  main_th = current;

  ready_q = sthread_new_queue();
  assert(ready_q != NULL);

  exiting = NULL;
/*</solution>*/  
}

sthread_t sthread_user_create(sthread_start_func_t start_routine, void *arg) {
/*<solution>*/    
  sthread_t sth;
  sth = (sthread_t)malloc(sizeof(struct _sthread));
  if (sth == NULL)
    return NULL;
  
  sth->start_routine = start_routine;
  sth->start_arg = arg;

  sth->saved_ctx = sthread_new_ctx(sthread_user_starter);
  if (sth->saved_ctx == NULL) {
    free(sth);
    return NULL;
  }

  printf("User create\n");
  sthread_enqueue(ready_q, sth);
  return sth;

#if 0  
/*</solution>*/
  return NULL;
/*<solution>*/
#endif
/*</solution>*/
}

void sthread_user_exit(void *ret) {
/*<solution>*/  
  exiting = sthread_user_current();
  sthread_user_schedule();
  abort();
/*</solution>*/    
}

void sthread_user_yield(void) {
/*<solution>*/    
  sthread_enqueue(ready_q, sthread_user_current());
  sthread_user_schedule();
/*</solution>*/  
}

/* Add any new part 1 functions here */

/*<solution>*/    
void sthread_user_starter(void) {
  /* This point is 1 of 2 places that we might be when we switch
   * contexts (the other is in sthread_user_schedule). */
  sthread_t sth;
  void *ret;

  printf("Starting user!\n");

  /* Handle the case where the previous thread exited, and
   * we switch to a brand new thread. */
  sthread_user_check_cleanup();
  
  sth = sthread_user_current();
  ret = sth->start_routine(sth->start_arg);
  sthread_user_exit(ret);
}

sthread_t sthread_user_current(void) {
  return current;
}

void sthread_user_schedule(void) {
  sthread_t next, tmp;
  next = sthread_user_next();
  tmp = current;
  current = next;
  printf("switching from 0x%x to 0x%x\n", (int)tmp->saved_ctx, (int)next->saved_ctx);

  sthread_switch(tmp->saved_ctx, next->saved_ctx);
  /* This point is 1 of 2 places that we might be when we switch
   * contexts (the other is in sthread_user_starter). */
  
  sthread_user_check_cleanup();
}

sthread_t sthread_user_next() {
  sthread_t next;
  next = sthread_dequeue(ready_q);
  if (next == NULL)
    exit(0);
  return next;
}

void sthread_user_check_cleanup(void) {
  if (exiting != NULL) {
    sthread_user_cleanup(exiting);
    exiting = NULL;
  }
}

void sthread_user_cleanup(sthread_t sth) {
  printf("cleaning 0x%x\n", (int)sth->saved_ctx);
  sthread_free_ctx(sth->saved_ctx);
  free(sth);
}

/* Add given thread to the ready queue. */
void sthread_user_ready(sthread_t ready_th) {
  sthread_enqueue(ready_q, ready_th);
}

/*</solution>*/    

/*********************************************************************/
/* Part 2: Synchronization Primitives                                */
/*********************************************************************/

struct _sthread_mutex {
  /* Fill in mutex data structure */
/*<solution>*/
  int locked;
  sthread_queue_t waiters;
/*</solution>*/
};

sthread_mutex_t sthread_user_mutex_init() {
/*<solution>*/
  sthread_mutex_t lock;
  lock = (sthread_mutex_t)malloc(sizeof(struct _sthread_mutex));
  assert(lock != NULL);
  lock->waiters = sthread_new_queue();
  lock->locked = 0;
  return lock;
#if 0
/*</solution>*/
  return NULL;
/*<solution>*/
#endif
/*</solution>*/
}

void sthread_user_mutex_free(sthread_mutex_t lock) {
/*<solution>*/
  assert(sthread_queue_is_empty(lock->waiters));
  sthread_free_queue(lock->waiters);
  free(lock);
/*</solution>*/  
}

void sthread_user_mutex_lock(sthread_mutex_t lock) {
/*<solution>*/  
  while (lock->locked) {
    sthread_enqueue(lock->waiters, sthread_user_current());
    sthread_user_schedule();
  }
  assert(!lock->locked);
  lock->locked = 1;
/*</solution>*/      
}

void sthread_user_mutex_unlock(sthread_mutex_t lock) {
/*<solution>*/
  assert(lock->locked);
  lock->locked = 0;
  if (!sthread_queue_is_empty(lock->waiters))
    sthread_user_ready(sthread_dequeue(lock->waiters));
/*</solution>*/      
}


struct _sthread_cond {
  /* Fill in condition variable structure */
/*<solution>*/  
  sthread_queue_t waiters;
/*</solution>*/  
};

sthread_cond_t sthread_user_cond_init(void) {
/*<solution>*/
  sthread_cond_t cond;
  cond = (sthread_cond_t)malloc(sizeof(struct _sthread_cond));
  assert(cond != NULL);
  cond->waiters = sthread_new_queue();
  return cond;
#if 0
/*</solution>*/
  return NULL;
/*<solution>*/
#endif
/*</solution>*/
}

void sthread_user_cond_free(sthread_cond_t cond) {
/*<solution>*/  
  assert(sthread_queue_is_empty(cond->waiters));
  sthread_free_queue(cond->waiters);
  free(cond);
/*</solution>*/  
}

void sthread_user_cond_signal(sthread_cond_t cond) {
/*<solution>*/  
  if (!sthread_queue_is_empty(cond->waiters))
    sthread_user_ready(sthread_dequeue(cond->waiters));
/*</solution>*/  
}

void sthread_user_cond_broadcast(sthread_cond_t cond) {
/*<solution>*/  
  while (!sthread_queue_is_empty(cond->waiters))
    sthread_user_ready(sthread_dequeue(cond->waiters));
/*</solution>*/  
}

void sthread_user_cond_wait(sthread_cond_t cond,
			    sthread_mutex_t lock) {
/*<solution>*/
  sthread_mutex_unlock(lock);
  sthread_enqueue(cond->waiters, sthread_user_current());
  sthread_user_schedule();
  sthread_mutex_lock(lock);
/*</solution>*/
}


